    /*********************************************************************
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    (C) Valerio Orlandini 2010. <valerio.orlandini@gmx.com>
    *********************************************************************/

#include "include/deathcrush.h"

inline float runRadical (float *in, float *factor) {
	unsigned short int isInPos;
    if ((*in) > 0) isInPos = 1;
    else {isInPos = 0; (*in) *= -1.0f;}
    (*in) = powf((*in), (1.0f / (*factor)));
    if (isInPos) return (*in);
    else return (-1.0f * (*in));
}

inline float runComp(float *in, float *threshold, float *rate) {
	float negthreshold = -1.0f * (*threshold);
	if (*in < *threshold && *in > negthreshold) (*in) = (*in);
	else {
		if ((*in) > (*threshold)) (*in) = (*in - (((*in) - (*threshold)) / (*rate)));
		if ((*in) < negthreshold) (*in) = (*in - (((*in) - negthreshold) / (*rate)));
	}
	return (*in);
}

inline float deathcrush(float in, float *input_amp, float *crusher_rate, float *crusher_wet, float *sine_wet, float *distort_root, float *distort_wet, float *compress_thr, float *compress_rate, float *compress_wet, float *noise_vol) {
	float output = in * (*input_amp);
	float noise = (( 2.0f * (float)rand()) / (float)RAND_MAX) - 1.0f;
	output = (output * (1.0f - (*crusher_wet))) + (((ceilf(output * (*crusher_rate))) / (*crusher_rate)) * (*crusher_wet));
	output = (output * (1.0f - (*sine_wet))) + (sinf(output * M_PI) * (*sine_wet));
	output = (output * (1.0f - (*distort_wet))) + (runRadical(&output, distort_root) * (*distort_wet));
	output = (output * (1.0f - (*compress_wet))) + (runComp(&output, compress_thr, compress_rate) * (*compress_wet)); 
	return (output + (noise * (*noise_vol)));
}


static LV2_Descriptor *deathDescriptor = NULL;

typedef struct {
	float *input_l;
	float *input_r;
	
	float *output_l;
	float *output_r;

	float *input_amp;
	
	float *crusher_rate;
	float *crusher_wet;
	
	float *sine_wet;
	
	float *distort_root;
	float *distort_wet;

	float *compress_thr;
	float *compress_rate;
	float *compress_wet;
		
	float *noise_vol;
	
	float *output_amp;
} Death;

static void cleanupDeath(LV2_Handle instance)
{
	free(instance);
}


static void connectPortDeath(LV2_Handle instance, uint32_t port, void *data)
{
	Death *plugin = (Death *)instance;
	
	switch (port) {
	case INPUT_1:
		plugin->input_l = data;
		break;
	case INPUT_2:
		plugin->input_r = data;
		break;
	case OUTPUT_1:
		plugin->output_l = data;
		break;
	case OUTPUT_2:
		plugin->output_r = data;
		break;
	case INPUT_AMP:
		plugin->input_amp = data;
		break;
	case CRUSHER_RATE:
		plugin->crusher_rate = data;
		break;
	case CRUSHER_WET:
		plugin->crusher_wet = data;
		break;
	case SINE_WET:
		plugin->sine_wet = data;
		break;
	case DISTORT_ROOT:
		plugin->distort_root = data;
		break;
	case DISTORT_WET:
		plugin->distort_wet = data;
		break;
	case COMPRESS_THR:
		plugin->compress_thr = data;
		break;
	case COMPRESS_RATE:
		plugin->compress_rate = data;
		break;
	case COMPRESS_WET:
		plugin->compress_wet = data;
		break;
	case NOISE_VOL:
		plugin->noise_vol = data;
		break;
	case OUTPUT_AMP:
		plugin->output_amp = data;
		break;
	}
}

static LV2_Handle instantiateDeath (const LV2_Descriptor * descriptor, double s_rate, const char * bundle_path,
	const LV2_Feature * const * host_features)
{
	Death *plugin_data = (Death *)malloc(sizeof(Death));
	return (LV2_Handle)plugin_data;
}

static void runDeath(LV2_Handle instance, uint32_t sample_count)
{
	Death *plugin_data = (Death *)instance;
    
    srand(time(NULL));
    
	const float * const inl = plugin_data->input_l;
	const float * const inr = plugin_data->input_r;
	float in_amp = *(plugin_data->input_amp) / 100.0f;

	float * const outl = plugin_data->output_l;
	float * const outr = plugin_data->output_r;
	float out_amp = *(plugin_data->output_amp) / 100.0f;
	
	float crush_rat = (501.0f - *(plugin_data->crusher_rate));
	float crush_wet = *(plugin_data->crusher_wet) / 100.0f;
	
	float sin_wet = *(plugin_data->sine_wet) / 100.0f;
	
	float dist_root = *(plugin_data->distort_root);
	float dist_wet = *(plugin_data->distort_wet) / 100.0f;
	
	float comp_thr = *(plugin_data->compress_thr) / 100.0f;
	float comp_rat = *(plugin_data->compress_rate) + 0.1f;
	float comp_wet = *(plugin_data->compress_wet) / 100.0f;
	
	float noise = *(plugin_data->noise_vol) / 100.0f;
	
	register uint32_t pos;
 
	for (pos = 0; pos < sample_count; pos++) {
		outl[pos] = deathcrush(inl[pos], &in_amp, &crush_rat, &crush_wet, &sin_wet, &dist_root, &dist_wet, &comp_thr, &comp_rat, &comp_wet, &noise) * out_amp;
		outr[pos] = deathcrush(inr[pos], &in_amp, &crush_rat, &crush_wet, &sin_wet, &dist_root, &dist_wet, &comp_thr, &comp_rat, &comp_wet, &noise) * out_amp;
	}
}

static void init()
{
	deathDescriptor = (LV2_Descriptor *)malloc(sizeof(LV2_Descriptor));

	deathDescriptor->URI = "http://interferon.drivehq.com/plugins/deathcrush/";
	deathDescriptor->activate = NULL;
	deathDescriptor->cleanup = cleanupDeath;
	deathDescriptor->connect_port = connectPortDeath;
	deathDescriptor->deactivate = NULL;
	deathDescriptor->instantiate = instantiateDeath;
	deathDescriptor->run = runDeath;
	deathDescriptor->extension_data = NULL;
}

LV2_SYMBOL_EXPORT
const LV2_Descriptor *lv2_descriptor(uint32_t index)
{
	if (!deathDescriptor) init();

	switch (index) {
	case 0:
		return deathDescriptor;
	default:
		return NULL;
	}
}